from mpi4py import MPI

import dolfinx
import ipyparallel as ipp
import numpy as np
import numpy.typing
import pytest

import adios4dolfinx


@pytest.fixture(scope="module")
def cluster():
    cluster = ipp.Cluster(engines="mpi", n=2)
    rc = cluster.start_and_connect_sync()
    yield rc
    cluster.stop_cluster_sync()


@pytest.fixture(scope="function")
def write_function(tmp_path):
    def _write_function(mesh, el, f, dtype, name="uh", append: bool = False) -> str:
        V = dolfinx.fem.functionspace(mesh, el)
        uh = dolfinx.fem.Function(V, dtype=dtype)
        uh.interpolate(f)
        uh.name = name
        el_hash = (
            V.element.signature()
            .replace(" ", "")
            .replace(",", "")
            .replace("(", "")
            .replace(")", "")
            .replace("[", "")
            .replace("]", "")
        )
        # Consistent tmp dir across processes
        f_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
        file_hash = f"{el_hash}_{np.dtype(dtype).name}"
        filename = f_path / f"mesh_{file_hash}.bp"
        if mesh.comm.size != 1:
            if not append:
                adios4dolfinx.write_mesh(filename, mesh)
            adios4dolfinx.write_function(filename, uh, time=0.0)
        else:
            if MPI.COMM_WORLD.rank == 0:
                if not append:
                    adios4dolfinx.write_mesh(filename, mesh)
                adios4dolfinx.write_function(filename, uh, time=0.0)

        return filename

    return _write_function


@pytest.fixture(scope="function")
def read_function():
    def _read_function(comm, el, f, path, dtype, name="uh"):
        engine = "BP4"
        mesh = adios4dolfinx.read_mesh(path, comm, engine, dolfinx.mesh.GhostMode.shared_facet)
        V = dolfinx.fem.functionspace(mesh, el)
        v = dolfinx.fem.Function(V, dtype=dtype)
        v.name = name
        adios4dolfinx.read_function(path, v, engine)
        v_ex = dolfinx.fem.Function(V, dtype=dtype)
        v_ex.interpolate(f)

        res = np.finfo(dtype).resolution
        np.testing.assert_allclose(v.x.array, v_ex.x.array, atol=10 * res, rtol=10 * res)

    return _read_function


@pytest.fixture(scope="function")
def get_dtype():
    def _get_dtype(in_dtype: np.dtype, is_complex: bool):
        dtype: numpy.typing.DTypeLike
        if in_dtype == np.float32:
            if is_complex:
                dtype = np.complex64
            else:
                dtype = np.float32
        elif in_dtype == np.float64:
            if is_complex:
                dtype = np.complex128
            else:
                dtype = np.float64
        else:
            raise ValueError("Unsuported dtype")
        return dtype

    return _get_dtype


@pytest.fixture(scope="function")
def write_function_time_dep(tmp_path):
    def _write_function_time_dep(mesh, el, f0, f1, t0, t1, dtype) -> str:
        V = dolfinx.fem.functionspace(mesh, el)
        uh = dolfinx.fem.Function(V, dtype=dtype)
        uh.interpolate(f0)
        el_hash = (
            V.element.signature()
            .replace(" ", "")
            .replace(",", "")
            .replace("(", "")
            .replace(")", "")
            .replace("[", "")
            .replace("]", "")
        )
        file_hash = f"{el_hash}_{np.dtype(dtype).name}"
        # Consistent tmp dir across processes
        f_path = MPI.COMM_WORLD.bcast(tmp_path, root=0)
        filename = f_path / f"mesh_{file_hash}.bp"
        if mesh.comm.size != 1:
            adios4dolfinx.write_mesh(filename, mesh)
            adios4dolfinx.write_function(filename, uh, time=t0)
            uh.interpolate(f1)
            adios4dolfinx.write_function(filename, uh, time=t1)

        else:
            if MPI.COMM_WORLD.rank == 0:
                adios4dolfinx.write_mesh(filename, mesh)
                adios4dolfinx.write_function(filename, uh, time=t0)
                uh.interpolate(f1)
                adios4dolfinx.write_function(filename, uh, time=t1)

        return filename

    return _write_function_time_dep


@pytest.fixture(scope="function")
def read_function_time_dep():
    def _read_function_time_dep(comm, el, f0, f1, t0, t1, path, dtype):
        engine = "BP4"
        mesh = adios4dolfinx.read_mesh(path, comm, engine, dolfinx.mesh.GhostMode.shared_facet)
        V = dolfinx.fem.functionspace(mesh, el)
        v = dolfinx.fem.Function(V, dtype=dtype)

        adios4dolfinx.read_function(path, v, engine, time=t1)
        v_ex = dolfinx.fem.Function(V, dtype=dtype)
        v_ex.interpolate(f1)

        res = np.finfo(dtype).resolution
        assert np.allclose(v.x.array, v_ex.x.array, atol=10 * res, rtol=10 * res)

        adios4dolfinx.read_function(path, v, engine, time=t0)
        v_ex = dolfinx.fem.Function(V, dtype=dtype)
        v_ex.interpolate(f0)

        res = np.finfo(dtype).resolution
        assert np.allclose(v.x.array, v_ex.x.array, atol=10 * res, rtol=10 * res)

    return _read_function_time_dep



from mpi4py import MPI

import dolfinx
import numpy as np
import pytest
import ufl

from adios4dolfinx import read_mesh, write_mesh
from adios4dolfinx.adios2_helpers import adios2


@pytest.mark.parametrize("encoder, suffix", [("BP4", ".bp"), ("HDF5", ".h5"), ("BP5", ".bp")])
@pytest.mark.parametrize(
    "ghost_mode", [dolfinx.mesh.GhostMode.shared_facet, dolfinx.mesh.GhostMode.none]
)
@pytest.mark.parametrize("store_partition", [True, False])
def test_mesh_read_writer(encoder, suffix, ghost_mode, tmp_path, store_partition):
    N = 7
    # Consistent tmp dir across processes
    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    file = fname / f"adios_mesh_{encoder}_{store_partition}"
    xdmf_file = fname / "xdmf_mesh_{encode}_{ghost_mode}_{store_partition}"
    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, N, N, N, ghost_mode=ghost_mode)

    write_mesh(file.with_suffix(suffix), mesh, encoder, store_partition_info=store_partition)
    mesh.comm.Barrier()
    with dolfinx.io.XDMFFile(mesh.comm, xdmf_file.with_suffix(".xdmf"), "w") as xdmf:
        xdmf.write_mesh(mesh)
    mesh.comm.Barrier()

    mesh_adios = read_mesh(
        file.with_suffix(suffix),
        MPI.COMM_WORLD,
        engine=encoder,
        ghost_mode=ghost_mode,
        read_from_partition=store_partition,
    )
    mesh_adios.comm.Barrier()
    if store_partition:

        def compute_distance_matrix(points_A, points_B, tol=1e-12):
            points_A_e = np.expand_dims(points_A, 1)
            points_B_e = np.expand_dims(points_B, 0)
            distances = np.sum(np.square(points_A_e - points_B_e), axis=2)
            return distances < tol

        cell_map = mesh.topology.index_map(mesh.topology.dim)
        new_cell_map = mesh_adios.topology.index_map(mesh_adios.topology.dim)
        assert cell_map.size_local == new_cell_map.size_local
        assert cell_map.num_ghosts == new_cell_map.num_ghosts
        mesh.topology.create_connectivity(mesh.topology.dim, mesh.topology.dim)
        midpoints = dolfinx.mesh.compute_midpoints(
            mesh,
            mesh.topology.dim,
            np.arange(cell_map.size_local + cell_map.num_ghosts, dtype=np.int32),
        )
        mesh_adios.topology.create_connectivity(mesh_adios.topology.dim, mesh_adios.topology.dim)
        new_midpoints = dolfinx.mesh.compute_midpoints(
            mesh_adios,
            mesh_adios.topology.dim,
            np.arange(new_cell_map.size_local + new_cell_map.num_ghosts, dtype=np.int32),
        )
        # Check that all points in owned by initial mesh is owned by the new mesh
        # (might be locally reordered)
        owned_distances = compute_distance_matrix(
            midpoints[: cell_map.size_local], new_midpoints[: new_cell_map.size_local]
        )
        np.testing.assert_allclose(np.sum(owned_distances, axis=1), 1)
        # Check that all points that are ghosted in original mesh is ghosted on the
        # same process in the new mesh
        ghost_distances = compute_distance_matrix(
            midpoints[cell_map.size_local :], new_midpoints[new_cell_map.size_local :]
        )
        np.testing.assert_allclose(np.sum(ghost_distances, axis=1), 1)

    mesh.comm.Barrier()

    with dolfinx.io.XDMFFile(mesh.comm, xdmf_file.with_suffix(".xdmf"), "r") as xdmf:
        mesh_xdmf = xdmf.read_mesh(ghost_mode=ghost_mode)

    for i in range(mesh.topology.dim + 1):
        mesh.topology.create_entities(i)
        mesh_xdmf.topology.create_entities(i)
        mesh_adios.topology.create_entities(i)
        assert (
            mesh_xdmf.topology.index_map(i).size_global
            == mesh_adios.topology.index_map(i).size_global
        )

    # Check that integration over different entities are consistent
    measures = (
        [ufl.ds, ufl.dx] if ghost_mode is dolfinx.mesh.GhostMode.none else [ufl.ds, ufl.dS, ufl.dx]
    )
    for measure in measures:
        c_adios = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh_adios)))
        c_ref = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh)))
        c_xdmf = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh_xdmf)))
        assert np.isclose(
            mesh_adios.comm.allreduce(c_adios, MPI.SUM),
            mesh.comm.allreduce(c_xdmf, MPI.SUM),
        )
        assert np.isclose(
            mesh_adios.comm.allreduce(c_adios, MPI.SUM),
            mesh.comm.allreduce(c_ref, MPI.SUM),
        )


@pytest.mark.parametrize("encoder, suffix", [("BP4", ".bp"), ("BP5", ".bp")])
@pytest.mark.parametrize(
    "ghost_mode", [dolfinx.mesh.GhostMode.shared_facet, dolfinx.mesh.GhostMode.none]
)
@pytest.mark.parametrize("store_partition", [True, False])
def test_timedep_mesh(encoder, suffix, ghost_mode, tmp_path, store_partition):
    # Currently unsupported, unclear why ("HDF5", ".h5"),
    N = 13
    # Consistent tmp dir across processes
    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    file = fname / f"adios_time_dep_mesh_{encoder}"
    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, N, N, N, ghost_mode=ghost_mode)

    def u(x):
        return np.asarray([x[0] + 0.1 * np.sin(x[1]), 0.2 * np.cos(x[1]), x[2]])

    write_mesh(
        file.with_suffix(suffix),
        mesh,
        encoder,
        mode=adios2.Mode.Write,
        time=0.0,
        store_partition_info=store_partition,
    )
    delta_x = u(mesh.geometry.x.T).T
    mesh.geometry.x[:] += delta_x
    write_mesh(file.with_suffix(suffix), mesh, encoder, mode=adios2.Mode.Append, time=3.0)
    mesh.geometry.x[:] -= delta_x

    mesh_first = read_mesh(
        file.with_suffix(suffix),
        MPI.COMM_WORLD,
        encoder,
        ghost_mode,
        time=0.0,
        read_from_partition=store_partition,
    )
    mesh_first.comm.Barrier()

    # Check that integration over different entities are consistent
    measures = [ufl.ds, ufl.dx]
    if ghost_mode == dolfinx.mesh.GhostMode.shared_facet:
        measures.append(ufl.dx)
    for measure in measures:
        c_adios = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh_first)))
        c_ref = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh)))
        assert np.isclose(
            mesh_first.comm.allreduce(c_adios, MPI.SUM),
            mesh.comm.allreduce(c_ref, MPI.SUM),
        )

    mesh.geometry.x[:] += delta_x
    mesh_second = read_mesh(
        file.with_suffix(suffix),
        MPI.COMM_WORLD,
        encoder,
        ghost_mode,
        time=3.0,
        read_from_partition=store_partition,
    )
    mesh_second.comm.Barrier()
    measures = [ufl.ds, ufl.dx]
    if ghost_mode == dolfinx.mesh.GhostMode.shared_facet:
        measures.append(ufl.dx)
    for measure in measures:
        c_adios = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh_second)))
        c_ref = dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * measure(domain=mesh)))
        assert np.isclose(
            mesh_second.comm.allreduce(c_adios, MPI.SUM),
            mesh.comm.allreduce(c_ref, MPI.SUM),
        )



# Copyright (C) 2023 Jørgen Schartum Dokken
#
# This file is part of adios4dolfinx
#
# SPDX-License-Identifier:    MIT


import pathlib

from mpi4py import MPI

import dolfinx
import numpy as np
import pytest
import ufl
from dolfinx.fem.petsc import LinearProblem

from adios4dolfinx import (
    read_function,
    read_function_from_legacy_h5,
    read_mesh,
    read_mesh_from_legacy_h5,
)


def test_legacy_mesh():
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy") / "mesh.h5").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")
    mesh = read_mesh_from_legacy_h5(filename=path, comm=comm, group="/mesh")
    assert mesh.topology.dim == 3
    volume = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.dx(domain=mesh))),
        op=MPI.SUM,
    )
    surface = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.ds(domain=mesh))),
        op=MPI.SUM,
    )
    assert np.isclose(volume, 1)
    assert np.isclose(surface, 6)

    mesh.topology.create_entities(mesh.topology.dim - 1)
    num_facets = mesh.topology.index_map(mesh.topology.dim - 1).size_global
    assert num_facets == 18


def test_read_legacy_mesh_from_checkpoint():
    comm = MPI.COMM_WORLD
    filename = (pathlib.Path("legacy") / "mesh_checkpoint.h5").absolute()
    if not filename.exists():
        pytest.skip(f"{filename} does not exist")
    mesh = read_mesh_from_legacy_h5(filename=filename, comm=comm, group="/Mesh/mesh")
    assert mesh.topology.dim == 3
    volume = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.dx(domain=mesh))),
        op=MPI.SUM,
    )
    surface = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.ds(domain=mesh))),
        op=MPI.SUM,
    )
    assert np.isclose(volume, 1)
    assert np.isclose(surface, 6)

    mesh.topology.create_entities(mesh.topology.dim - 1)
    num_facets = mesh.topology.index_map(mesh.topology.dim - 1).size_global
    assert num_facets == 18


def test_legacy_function():
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy") / "mesh.h5").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")
    mesh = read_mesh_from_legacy_h5(path, comm, "/mesh")
    V = dolfinx.fem.functionspace(mesh, ("DG", 2))
    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    a = ufl.inner(u, v) * ufl.dx
    x = ufl.SpatialCoordinate(mesh)
    f = ufl.conditional(ufl.gt(x[0], 0.5), x[1], 2 * x[0])
    L = ufl.inner(f, v) * ufl.dx

    uh = dolfinx.fem.Function(V)
    problem = LinearProblem(a, L, [], uh, petsc_options={"ksp_type": "preonly", "pc_type": "lu"})
    problem.solve()

    u_in = dolfinx.fem.Function(V)
    read_function_from_legacy_h5(path, mesh.comm, u_in, group="v")
    np.testing.assert_allclose(uh.x.array, u_in.x.array, atol=1e-14)

    W = dolfinx.fem.functionspace(mesh, ("DG", 2, (mesh.geometry.dim,)))
    wh = dolfinx.fem.Function(W)
    wh.interpolate(lambda x: (x[0], 3 * x[2], 7 * x[1]))
    w_in = dolfinx.fem.Function(W)

    read_function_from_legacy_h5(path, mesh.comm, w_in, group="w")

    np.testing.assert_allclose(wh.x.array, w_in.x.array, atol=1e-14)


def test_read_legacy_function_from_checkpoint():
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy") / "mesh_checkpoint.h5").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")
    mesh = read_mesh_from_legacy_h5(path, comm, "/Mesh/mesh")

    V = dolfinx.fem.functionspace(mesh, ("DG", 2))
    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    a = ufl.inner(u, v) * ufl.dx
    x = ufl.SpatialCoordinate(mesh)
    f = ufl.conditional(ufl.gt(x[0], 0.5), x[1], 2 * x[0])
    L = ufl.inner(f, v) * ufl.dx

    uh = dolfinx.fem.Function(V)
    problem = LinearProblem(a, L, [], uh, petsc_options={"ksp_type": "preonly", "pc_type": "lu"})
    problem.solve()

    u_in = dolfinx.fem.Function(V)
    read_function_from_legacy_h5(path, mesh.comm, u_in, group="v", step=0)
    assert np.allclose(uh.x.array, u_in.x.array)

    # Check second step
    uh.interpolate(lambda x: x[0])
    read_function_from_legacy_h5(path, mesh.comm, u_in, group="v", step=1)
    assert np.allclose(uh.x.array, u_in.x.array)

    W = dolfinx.fem.functionspace(mesh, ("DG", 2, (mesh.geometry.dim,)))
    wh = dolfinx.fem.Function(W)
    wh.interpolate(lambda x: (x[0], 3 * x[2], 7 * x[1]))
    w_in = dolfinx.fem.Function(W)

    read_function_from_legacy_h5(path, mesh.comm, w_in, group="w", step=0)
    np.testing.assert_allclose(wh.x.array, w_in.x.array, atol=1e-14)

    wh.interpolate(lambda x: np.vstack((x[0], 0 * x[0], x[1])))
    read_function_from_legacy_h5(path, mesh.comm, w_in, group="w", step=1)
    np.testing.assert_allclose(wh.x.array, w_in.x.array, atol=1e-14)


def test_adios4dolfinx_legacy():
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy_checkpoint") / "adios4dolfinx_checkpoint.bp").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")

    el = ("N1curl", 3)
    mesh = read_mesh(path, comm, "BP4", dolfinx.mesh.GhostMode.shared_facet, legacy=True)

    def f(x):
        values = np.zeros((2, x.shape[1]), dtype=np.float64)
        values[0] = x[0]
        values[1] = -x[1]
        return values

    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    read_function(path, u, engine="BP4", legacy=True)

    u_ex = dolfinx.fem.Function(V)
    u_ex.interpolate(f)

    np.testing.assert_allclose(u.x.array, u_ex.x.array, atol=1e-14)



# Copyright (C) 2024 Jørgen Schartum Dokken
#
# This file is part of adios4dolfinx
#
# SPDX-License-Identifier:    MIT

from __future__ import annotations

import typing
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

"""Internal library classes for storing mesh and function data"""
__all__ = ["MeshData", "FunctionData"]


@dataclass
class MeshData:
    # 2 dimensional array of node coordinates
    local_geometry: npt.NDArray[np.floating]
    local_geometry_pos: tuple[int, int]  # Insert range on current process for geometry nodes
    num_nodes_global: int  # Number of nodes in global geometry array

    local_topology: npt.NDArray[np.int64]  # 2 dimensional connecitivty array for mesh topology
    # Insert range on current process for topology
    local_topology_pos: tuple[int, int]
    num_cells_global: int  # NUmber of cells in global topology

    cell_type: str
    degree: int
    lagrange_variant: int

    # Partitioning_information
    store_partition: bool
    partition_processes: typing.Optional[int]  # Number of processes in partition
    ownership_array: typing.Optional[npt.NDArray[np.int32]]  # Ownership array for cells
    ownership_offset: typing.Optional[npt.NDArray[np.int32]]  # Ownership offset for cells
    partition_range: typing.Optional[
        tuple[int, int]
    ]  # Local insert position for partitioning information
    partition_global: typing.Optional[int]


@dataclass
class FunctionData:
    cell_permutations: npt.NDArray[np.uint32]  # Cell permutations for dofmap
    local_cell_range: tuple[int, int]  # Range of cells on current process
    num_cells_global: int  # Number of cells in global topology
    dofmap_array: npt.NDArray[np.int64]  # Local function dofmap (using global indices)
    dofmap_offsets: npt.NDArray[np.int64]  # Global dofmap offsets
    dofmap_range: tuple[int, int]  # Range of dofmap on current process
    global_dofs_in_dofmap: int  # Number of entries in global dofmap
    values: npt.NDArray[np.floating]  # Local function values
    dof_range: tuple[int, int]  # Range of local function values
    num_dofs_global: int  # Number of global function values
    name: str  # Name of function
